/*
 * Copyright (c) 2001-2004 Greg Becker
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: dits.c,v 1.8 2004/04/05 02:53:27 greg Exp $
 */

/* dits - Data Integrity Test Suite
 *
 * dits is a tool for validating disk drivers, file systems, and
 * lock managers.  More than an exerciser, dits is able to verify
 * that no detectable data integrity errors have occurred.
 *
 * The tool does its work in three mutually exclusive phases:
 *
 * (1) Init - Write a unique ID to each sector in the test bed.
 *
 * (2) Test - Continuously select arbitrary non-overlapping ranges of
 * sectors and swap them.  File range locking is used to ensure mutual
 * exclusion amongst concurrently executing processes working in the
 * same test bed.
 *
 * (3) Check - Verify that each and every unique ID's written in
 * phase are intact and neither missing nor duplicated.
 */

static char rcsid[] = "$Id: dits.c,v 1.8 2004/04/05 02:53:27 greg Exp $";

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#if STDC_HEADERS
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#else
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#endif

#if HAVE_STRING_H
#if !STDC_HEADERS && HAVE_MEMORY_H
#include <memory.h>
#endif
#include <string.h>
#endif

#if HAVE_STRINGS_H
#include <strings.h>
#endif

#if HAVE_INTTYPES_H
#include <inttypes.h>
#else
#if HAVE_STDINT_H
#include <stdint.h>
#endif
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#include <assert.h>
#include <signal.h>
#include <sysexits.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/mman.h>

#if HAVE_AIO
#include <aio.h>
#endif /* HAVE_AIO */

#define NDEBUG

#include "dits.h"


/* Minimum number of sectors to manage read/write per syscall during
 * init() and check().
 */
#define MINSWAPSECTORS		2048	/* Must be a power of 2 !!! */


/* All the test process share a table of info records into which they
 * record their individual current active range locks.  If a test
 * process dies, its file locks are released automatically by the
 * OS, but its range locks in the info table persist, which is a
 * good thing because it keeps other test processes from using those
 * locked ranges until the parent has a chance to roll back the
 * transaction and release the locks.
 */
typedef struct {
	int locked;			/* Range locks are valid */
	pid_t pid;			/* Pid of test process */
	struct flock src;		/* Source range lock */
	struct flock dst;		/* Destination range lock */
} info_t;


/* The only type of repair that is possible is recovery from a
 * swap that has only partially taken place.  E.g., while swapping
 * blocks A and B, after reading them both into core we write A's
 * data to block B and then crash.  On the death of a child, the
 * parent scans the blocks of the child's locked ranges and
 * generates a defect list of all the id's found and the address
 * at which they were found (x and y).  So, x[id] contains the
 * sector number of the record with the id.  If an id is encountered
 * a second time, its address is stored in y[id].  We can then
 * simply scan y for valid entries and roll them all back.
 */
typedef struct {
	int64_t nsectors;		/* Number of sectors in x&y lists */
	int64_t inplace;		/* # whose id matched block */
	int64_t *x;			/* First occurrence list */
	int64_t *y;			/* Second occurrence list */
} defect_t;


#if HAVE_AIO
#ifdef _AIO_AIX_SOURCE

/* Legacy aio on aix.
 */
#define AIO_ERROR(aiocb)	aio_error((aiocb)->aio_handle)
#define AIO_SUSPEND(ap, apsz)	aio_suspend((apsz), (ap))

#else

/* POSIX aio.
 */
#define AIO_ERROR(aiocb)	aio_error((aiocb))
#define AIO_SUSPEND(ap, apsz)	\
	aio_suspend((const struct aiocb **)(ap), (apsz), (struct timespec *)0)

#endif /* _AIO_AIX_SOURCE */


/* xaio is used to track aio reads and writes.  Basically, every
 * swap involves two xaio records which are tied together via
 * the "sibling" link.  The records are initialized and dispatched.
 * Rather than wait on the request, the caller continues,
 * generating more requests, periodically checking and dealing
 * with requests that finished.
 */
typedef struct xaio {
	struct aiocb aio;		/* Must be first field!!! */
	struct flock lk;
	sector_t *sb;
	int nrgns;
	int done;
	int idx;
	struct xaio *sibling;
	int read;
	struct xaio *next;		/* xaio free queue list node */
	u_long magic;
} xaio_t;

static xaio_t *xaiohead = 0;
int getidx(int n, xaio_t **ap, int *pidx);
int chkoverlap(int n, xaio_t **ap, xaio_t *tgt);
int getaiobuf(int nrgns, int fd, xaio_t **x1, xaio_t **x2);
void freeaiobuf(xaio_t *x1, xaio_t *x2);
void testaio(info_t *infobase, int fdInfo,
	     int fdLock, int fdPart, int64_t nsectors);

typedef struct aiocb		*aiocbptr_t;

#else

typedef void 			*aiocbptr_t;

#endif /* HAVE_AIO */


/* Variables set by the option parser.
 * f{varname} - flag, true if option was given
 * arg{varname} - optarg argument
 */
int fExec = 1;
int verbosity = 0;
int fAIO, fCheck, fConcurrency, fDump, fOSeek, fInit, fRepair, fTest;
int fHelp, fVersion;

char *excludes[256];
char given[256];

int64_t maxswapsectors = MINSWAPSECTORS;
int maxiterations = 0;
u_long maxblksps = 0;
int64_t oseek = 0;
int naioreqs = 2;
int nprocs = 4;
char *progname;
int fSigAlarm;
int fSig;

char state[256];			/* For initstate() */


void init(char *partition, int64_t sectors);
void check(char *partition);
void teststart(char *partition);
void test(info_t *infobase, int fdInfo,
	  int fdLock, int fdPart, int64_t nsectors);
void dump(char *partition, int64_t start, int64_t stop);
void chkrgn(sector_t *sb, int64_t offset, size_t len, defect_t *d,
	    xx_crc32_t *signature);
void saverollback(sector_t *x, sector_t *y, size_t nrgns);
void dumprgn(sector_t *sb, int64_t offset, size_t len, int verbosity);
void updatergn(sector_t *sb, int64_t src, int64_t dst,
	       size_t len, pid_t pid, time_t time);
int repair(int fdPart, defect_t *d);
int cleanlocks(info_t *victim, int fdInfo, int fdPart, int64_t nsectors);
RETSIGTYPE sigHandler(int sig);
RETSIGTYPE sigAlarm(int sig);

extern void getrgn(int fd, sector_t *sb, int64_t offset, size_t len,
		   aiocbptr_t aio);

extern void putrgn(int fd, sector_t *sb, int64_t offset, size_t len,
		   aiocbptr_t aio);

void
usage()
{
	printf("usage: %s -i [-s swap ] [-v] partition nsectors\n",
	       progname);
	printf("usage: %s -t [-a naioreqs] [-b blksps] [-nv] [-C nprocs] "
		"[-I iter] [-s swap] partition\n", progname);
	printf("usage: %s -c [-rv] [-s swap] partition\n", progname);
	printf("usage: %s -d [-v] partition start [stop]\n", progname);
	printf("usage: %s -h\n", progname);
	printf("usage: %s -V\n", progname);

	printf("-a nreqs     use aio [%d < nreqs] (default: %d)\n",
	       2, naioreqs);
	printf("-b blksps    set max blocks/sec\n");
	printf("-C nprocs    set concurrency (default: %d)\n", nprocs);
	printf("-d           dump the given range of sectors\n");
	printf("-c           check\n");
	printf("-h           show this help list\n");
	printf("-I iter      max iterations\n");
	printf("-i           initialize test bed\n");
	printf("-n           don't really perform the swap\n");
	printf("-o nsectors  skip first nsectors in partition\n");
	printf("-r           try to repair a corrupted test bed\n");
	printf("-s swap      max sectors to swap (default: %ld)\n",
	       (long)maxswapsectors);
	printf("-t           test\n");
	printf("-V           print version\n");
	printf("-v           be verbose\n");
	printf("partition    a disk partition, volume, file, etc, ...\n");
	printf("nsectors     number of sectors in test bed [nsectors >= %d]\n",
	       MINSWAPSECTORS);
}


int
chkexcludes(int c)
{
	char *msg = (char *)0;
	char *pc;
	int i;

	/* First, check to see if any option already given excludes
	 * this option.
	 */
	for (i = 0; i < 256; ++i) {
		if (given[i]) {
			for (pc = excludes[i]; pc && *pc; ++pc) {
				if (*pc == (char)c) {
					msg = "excludes";
					goto done;
				}
			}
		}
	}

	/* Next, check to see if this option excludes any already given.
	 */
	for (pc = excludes[c]; pc && *pc; ++pc) {
		if (given[(int)*pc]) {
			msg = "is excluded by";
			i = *pc;
			goto done;
		}
	}

 done:
	if (msg) {
		eprintf("%s: option `-%c' %s `-%c', use `-h' for help\n",
			progname, (char)i, msg, (char)c);
		return i;
	}

	return 0;
}


int
main(int argc, char **argv)
{
	int nargsexpected;
	int noptionalargs;
	char *envopts;
	int c;

	assert(sizeof(sector_t) == 512);

#if HAVE_AIO
	assert(sizeof(xaio_t) < sizeof(sector_t));
#endif /* HAVE_AIO */

	progname = strrchr(argv[0], (int)'/');
	progname = (progname ? progname+1 : argv[0]);

	(void)initstate((u_long)time((time_t *)0), state, sizeof(state));
	xxcrc32_init();

	/* TODO: Get options from the environment.
	 */
	envopts = getenv("DITS");
	if (envopts) {
		eprintf("DITS=%s ignored\n", envopts);
	}

	/* Initialize the option exlusion tables.
	 */
	excludes['c'] = "abCIhinstV";
	excludes['d'] = "abCcIhinorstV";
	excludes['h'] = "abCcdIinorstVv";
	excludes['i'] = "abCcdIhnrstV";
	excludes['r'] = "abCdhIirstV";
	excludes['t'] = "cdhinrV";
	excludes['V'] = "abCcdIhinorstv";

	nargsexpected = 0;
	noptionalargs = 0;

	while (-1 != (c = getopt(argc, argv, ":a:b:C:cdhI:ino:rs:tVv"))) {
		if (chkexcludes(c)) {
			exit(EX_USAGE);
		}
		given[c] = c;

		switch (c) {
		case 'a':
#if HAVE_AIO
			fAIO = !0;
			naioreqs = atoi(optarg);
			if (naioreqs < 2) {
				eprintf("nreqs must be [2 < nreqs]\n");
				exit(EX_USAGE);
			}
			if (naioreqs & 0x0001) {
				eprintf("nreqs must be a multiple of 2\n");
				exit(EX_USAGE);
			}

			/* TODO: For now we set nprocs to 1 because mulitple
			 * procs using aio don't play well together.
			 */
			if (fConcurrency && (nprocs > 1)) {
				dprintf(0, "Use of aio not compatible with "
					"more than one process.  Using -C1\n");
			}
			nprocs = 1;
#else
			eprintf("Not built with -DHAVE_AIO\n");
			exit(EX_USAGE);
#endif /* HAVE_AIO */
			break;

		case 'b':
			maxblksps = strtoul(optarg, (char **)0, 0);
			if (maxblksps < 0) {
				eprintf("max blks/sec must be >= 0\n");
				exit(EX_USAGE);
			}
			break;

		case 'C':
			fConcurrency = !0;
			nprocs = atoi(optarg);
			if (nprocs < 1) {
				eprintf("nprocs must be > 0\n");
				exit(EX_USAGE);
			}

			if (fAIO && (nprocs > 1)) {
				dprintf(0, "Use of aio not compatible with "
					"more than one process.  Using -C1\n");
				nprocs = 1;
			}
			break;

		case 'c':
			fCheck = !0;
			nargsexpected = 1;
			break;

		case 'd':
			fDump = !0;
			nargsexpected = 2;
			noptionalargs = 1;
			break;

		case 'h':
			fHelp = !0;
			break;

		case 'I':
			maxiterations = atol(optarg);
			if (maxiterations < 0) {
				eprintf("max iter must be >= 0\n");
				exit(EX_USAGE);
			}
			break;

		case 'i':
			fInit = !0;
			nargsexpected = 2;
			break;
			
		case 'n':
			fExec = 0;
			break;

		case 'o':
			fOSeek = !0;
			oseek = atol(optarg);
			if (oseek < 0) {
				eprintf("oseek must be >= 0\n");
				exit(EX_USAGE);
			}
			break;

		case 'r':
			fRepair = !0;
			break;

		case 's':
			maxswapsectors = atol(optarg);
			if (maxswapsectors < 0) {
				eprintf("max swap sectors must be > 0\n");
				exit(EX_USAGE);
			}
			break;

		case 't':
			fTest = !0;
			nargsexpected = 1;
			break;

		case 'V':
			fVersion = !0;
			break;

		case 'v':
			++verbosity;
			break;

		case '?':
			eprintf("invalid option `-%c'\n", optopt);
			exit(EX_USAGE);

		default:
			eprintf("option `-%c' requires a parameter\n", optopt);
			exit(EX_USAGE);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc < nargsexpected) {
		eprintf("mandatory arguments required, use `-h' for help\n");
		exit(EX_USAGE);
	} else if (argc > nargsexpected + noptionalargs) {
		eprintf("extraneous arguments detected, use `-h' for help\n");
		exit(EX_USAGE);
	}

	if (fInit) {
		int64_t nsectors;

		nsectors = atol(argv[1]);
		if (nsectors < MINSWAPSECTORS) {
			eprintf("nsectors (%s) must be >= %ld\n",
				argv[1], (long)MINSWAPSECTORS);
			exit(EX_USAGE);
		}

		/* Forcing nsectors to be a multiple of MINSWAPSECTORS makes it
		 * so that we can deal with the whole test bed in
		 * MINSWAPSECTORS sized chunks and eliminate code that deals
		 * with fractional segments.
		 */
		if (nsectors != (nsectors & ~(MINSWAPSECTORS - 1))) {
			dprintf(0, "reducing nsectors from %ld to %ld so that "
				"it is an integral multiple of "
				"MINSWAPSECTORS\n", (long)nsectors,
				(long)(nsectors & ~(MINSWAPSECTORS - 1)));
		}
		nsectors &= ~(MINSWAPSECTORS - 1);

		init(argv[0], nsectors);
	} else if (fTest) {
		teststart(argv[0]);
	} else if (fCheck) {
		check(argv[0]);
	} else if (fDump) {
		int64_t start, stop;

		start = atol(argv[1]);
		if (start < 0) {
			eprintf("start must be >= 0\n");
			exit(EX_USAGE);
		}
		stop = start + 1;

		if (argc > 2) {
			stop = atol(argv[2]);
			if (stop < start) {
				eprintf("stop must be >= start (%lld)\n",
					start);
				exit(EX_USAGE);
			}
		}

		dump(argv[0], start, stop);
	} else if (fHelp) {
		usage();
	} else if (fVersion) {
		printf("%s\n", VERSION);
#if 0
#ifdef _AIO_AIX_SOURCE
		printf("Using legacy aio.\n");
#else
		printf("Using POSIX aio.\n");
#endif /* _AIO_AIX_SOURCE */
#endif
	} else {
		eprintf("one of [-cdhitV] required, use `-h' for help\n");
		exit(EX_USAGE);
	}

	return 0;
}


/* Write a unique record to each sector of the given partition.
 */
void
init(char *partition, int64_t nsectors)
{
	char infofile[128];
	char *src, *dst;
	sector_t *sb;
	size_t sbsz;
	time_t now;
	int fdPart;
	int i, j;

	fdPart = open(partition, O_RDWR);
	if (-1 == fdPart) {
		eprintf("open(%s): %s\n", partition, strerror(errno));
		exit(EX_OSERR);
	}

	if (-1 == lseek(fdPart, oseek * sizeof(*sb), SEEK_SET)) {
		eprintf("init: lseek: %s\n", strerror(errno));
		exit(EX_OSERR);
	}

	dprintf(1, "nsectors=%lld maxswapsectors=%lld MINSWAPSECTORS=%d\n",
		nsectors, maxswapsectors, MINSWAPSECTORS);
	assert(nsectors == (nsectors & ~(MINSWAPSECTORS - 1)));

	sbsz = MINSWAPSECTORS * sizeof(*sb);
	sb = malloc(sbsz);
	if (!sb) {
		eprintf("malloc(%d) failed\n", sbsz);
		exit(EX_OSERR);
	}
	bzero(sb, sbsz);

	/* We buffer a "region" of changes and write them out
	 * in one call to write().
	 */
	now = time((time_t *)0);
	for (i = 0; i < nsectors; i += MINSWAPSECTORS) {
		for (j = 0; j < MINSWAPSECTORS; ++j) {
			record_t *r = &sb[j].rec[0];

			/* Initialize the invariant fields.
			 */
			r->id = i + j + oseek;
			r->nsectors = nsectors;

			r[1].id = -1;
		}

		bcopy(state, sb->payload, sizeof(sb->payload));

		dprintf(2, "init: initializing sectors [%lld - %lld]\n",
			i + oseek, i + oseek + MINSWAPSECTORS - 1);

		updatergn(sb, i + oseek, i + oseek, MINSWAPSECTORS, -1, now);
		putrgn(fdPart, sb, i + oseek, MINSWAPSECTORS, (aiocbptr_t)0);
	}

	free(sb);
	(void)close(fdPart);

	/* Remove the info file if it exists.
	 */
	sprintf(infofile, "/var/tmp/%s-", progname);
	dst = infofile + strlen(infofile);
	for (src = partition; *src != '\000'; ++src) {
		*dst++ = (*src == '/') ? '-' : *src;
	}
	*dst = '\000';

	dprintf(2, "removing infofile %s\n", infofile);
	(void)unlink(infofile);

}


/* Allocate a defect table.  This can be quite large for a large test bed.
 */
defect_t *
defect_alloc(int64_t nsectors)
{
	defect_t *d;

	d = malloc(sizeof(*d));
	if (!d) {
		eprintf("defect_alloc: malloc(%ld) failed\n", sizeof(*d));
		return (defect_t *)0;
	}
	bzero((char *)d, sizeof(*d));

	d->x = malloc(sizeof(*d->x) * nsectors);
	if (!d->x) {
		eprintf("defect_alloc: malloc(%lld) failed\n",
			sizeof(*d->x) * nsectors);
		free(d);
		return (defect_t *)0;
	}
	memset(d->x, -1, sizeof(*d->x) * nsectors);

	d->y = malloc(sizeof(*d->y) * nsectors);
	if (!d->y) {
		eprintf("defect_alloc: malloc(%lld) failed\n",
			sizeof(*d->y) * nsectors);
		free(d->x);
		return (defect_t *)0;
	}
	memset(d->y, -1, sizeof(*d->y) * nsectors);

	d->nsectors = nsectors;

	return d;
}


void
defect_free(defect_t *d)
{
	free(d->x);
	free(d->y);
	free(d);
}


/* Check that all the data written by init() is intact and that there
 * are no missing nor duplicated records.
 */
void
check(char *partition)
{
	xx_crc32_t signature;
	char infofile[128];
	int64_t nsectors;
	char *src, *dst;
	sector_t *sb;
	defect_t *d;
	size_t sbsz;
	int64_t rgn;
	char *pbase;
	int fdPart;
	int64_t i;
	int nerrs;

	fdPart = open(partition, (fRepair ? O_RDWR : O_RDONLY));
	if (-1 == fdPart) {
		eprintf("check(%s): open: %s\n", partition, strerror(errno));
		exit(EX_OSERR);
	}

	/* Get the base name of the partition.
	 */
	pbase = strrchr(partition, '/');
	if (!pbase) {
		pbase = partition;
	}

	sbsz = MINSWAPSECTORS * sizeof(*sb);
	sb = malloc(sbsz);
	if (!sb) {
		eprintf("check(%s): malloc(%ld) failed\n", pbase, sbsz);
		exit(EX_OSERR);
	}

	/* Get the first record which tells how many records are
	 * in the test bed.
	 */
	getrgn(fdPart, sb, oseek, (size_t)1, (aiocbptr_t)0);
	chkrgn(sb, oseek, (size_t)1, (defect_t *)0, (xx_crc32_t *)0);
	nsectors = sb->rec[0].nsectors;

	dprintf(2, "check(%s): nsectors=%lld maxswapsectors=%lld oseek=%lld\n",
		pbase, nsectors, maxswapsectors, oseek);

	d = defect_alloc(nsectors + oseek);
	if (!d) {
		exit(EX_OSERR);
	}

	/* Iteratively read in chunks of the test bed so that chkrgn
	 * can populate the defect table for each chunk.
	 */
	signature = 0;
	for (rgn = 0; rgn < nsectors; rgn += MINSWAPSECTORS) {
		dprintf(3, "check(%s): reading sectors [%lld-%lld] sig=0x%X\n",
			pbase, (long long)(rgn + oseek),
			(long long)(rgn + MINSWAPSECTORS + oseek), signature);
		getrgn(fdPart, sb, rgn + oseek, MINSWAPSECTORS, (aiocbptr_t)0);
		chkrgn(sb, rgn + oseek, MINSWAPSECTORS, d, &signature);
	}

	if (fRepair) {
		repair(fdPart, d);
	}

	/* Check that each record id maps to one and only one sector.
	 */
	nerrs = 0;
	dprintf(3, "check: checking record id mappings...\n");
	for (i = oseek; i < nsectors + oseek; ++i) {
		if ((d->x[i] == -1) || (d->y[i] != -1)) {
			eprintf("check(%s): *** DATA INTEGRITY ERROR *** "
				"sector %lld (%lld %lld)\n",
				pbase, i, d->x[i], d->y[i]);
			++nerrs;
		}
	}

	if (nerrs) {
		eprintf("check: data integrity errors: %d\n", nerrs);
		exit(EX_DATAERR);
	}

	dprintf(1, "check(%s): nsectors=%lld inplace=%lld (%3.2f%%) "
		"signature=0x%X\n",
		pbase, d->nsectors, d->inplace,
		(d->inplace * 100.0)/d->nsectors, signature);

	/* Remove the info file if it exists.
	 */
	sprintf(infofile, "/var/tmp/%s-", progname);
	dst = infofile + strlen(infofile);
	for (src = partition; *src != '\000'; ++src) {
		*dst++ = (*src == '/') ? '-' : *src;
	}
	*dst = '\000';

	dprintf(2, "removing infofile %s\n", infofile);
	(void)unlink(infofile);

	defect_free(d);
	free(sb);
	(void)close(fdPart);
}


/* Scan the defect list for duplicates.  The sector address of the first
 * occurrence of an id is in d->x[id], and the second occurrence, if any,
 * is in d->y[id].
 */
int
repair(int fdPart, defect_t *d)
{
	int64_t nrepaired;
	sector_t x, y;
	int64_t i;

	nrepaired = 0;
	for (i = 0; i < d->nsectors; ++i) {
		if (d->y[i] == -1) {
			continue;
		}

		getrgn(fdPart, &x, d->x[i], (size_t)1, (aiocbptr_t)0);
		dumprgn(&x, d->x[i], (size_t)1, 3);

		getrgn(fdPart, &y, d->y[i], (size_t)1, (aiocbptr_t)0);
		dumprgn(&y, d->y[i], (size_t)1, 3);

		/* If the id's don't match, then these sectors were not
		 * involved in the same interrupted swap transaction.  The
		 * implications are that this record has occurred more than
		 * twice, meaning there are serious problems somewhere.
		 */
		if (x.rec[0].id != y.rec[0].id) {
			eprintf("repair: id mismatch: sector/id: "
				"%lld/%lld %lld/%lld\n",
				d->x[i], x.rec[0].id,
				d->y[i], y.rec[0].id);
			continue;
		}

		dprintf(2, "repair: repairing sectors %lld %lld\n",
			d->x[i], d->y[i]);

		/* Roll back the problem sector.
		 */
		if (d->x[i] < d->y[i]) {
			x.rec[0] = x.rec[1];
			bzero(&x.rec[1], sizeof(x.rec[1]));
			x.rec[1].id = -1;
			putrgn(fdPart, &x, d->x[i], 1, (aiocbptr_t)0);
			assert(d->x[x.rec[0].id] == -1);
			d->x[x.rec[0].id] = d->x[i];
		} else {
			y.rec[0] = y.rec[1];
			bzero(&y.rec[1], sizeof(y.rec[1]));
			y.rec[1].id = -1;
			putrgn(fdPart, &y, d->y[i], 1, (aiocbptr_t)0);
			assert(d->x[y.rec[0].id] == -1);
			d->x[y.rec[0].id] = d->y[i];
		}
		d->y[i] = -1;
		++nrepaired;
	}

	dprintf(1, "repair: repaired %lld sectors\n", nrepaired);

	return 0;
}


/* Check that all the data written by init() is intact and that there
 * are no missing nor duplicated records.
 */
void
dump(char *partition, int64_t start, int64_t stop)
{
	char infofile[128];
	int64_t nsectors;
	char *src, *dst;
	sector_t *sb;
	size_t sbsz;
	int64_t rgn;
	int fdPart;
	int64_t i;

	assert(start >= 0);
	assert(stop >= start);

	fdPart = open(partition, O_RDONLY);
	if (-1 == fdPart) {
		eprintf("check: open(%s): %s\n", partition, strerror(errno));
		exit(EX_OSERR);
	}

	sbsz = MINSWAPSECTORS * sizeof(*sb);
	sb = malloc(sbsz);
	if (!sb) {
		eprintf("check: malloc(%ld) failed\n", sbsz);
		exit(EX_OSERR);
	}

	/* Get the first record which tells how many records are
	 * in the test bed.
	 */
	getrgn(fdPart, sb, oseek, (size_t)1, (aiocbptr_t)0);
	chkrgn(sb, oseek, (size_t)1, (defect_t *)0, (xx_crc32_t *)0);
	nsectors = sb->rec[0].nsectors;

	dprintf(2, "dump: nsectors=%lld maxswapsectors=%lld oseek=%lld\n",
		nsectors, maxswapsectors, oseek);

	/* Iteratively read in chunks of the test bed so that chkrgn
	 * can populate the defect table for each chunk.
	 */
	for (rgn = start; rgn < stop; rgn += MINSWAPSECTORS) {
		dprintf(3, "check: reading sectors [%lld-%lld]\n",
			rgn, rgn + MINSWAPSECTORS);
		getrgn(fdPart, sb, rgn,
		       stop - start, (aiocbptr_t)0);
		dumprgn(sb, rgn, stop - start, 0);
	}

	free(sb);
	(void)close(fdPart);
}


void
dumprgn(sector_t *sb, int64_t offset, size_t nrgns, int verbosity)
{
	char msg[256];
	int i, j;

	snprintf(msg, sizeof(msg), "  sector rec       id   pid   update "
			 "[     src      dst   len]      crc\n");
	dprintf(verbosity, "%s", msg);

	for (i = 0; i < nrgns; ++i, ++sb) {
		for (j = 0; j < 2; ++j) {
			snprintf(msg, sizeof(msg), "%8lld %3d "
				 "%8lld %5lld %8llX "
				 "[%8lld %08lld %5d] %08X\n",
				 (long long)offset+i, j,
				 (long long)sb->rec[j].id,
				 sb->rec[j].pid, sb->rec[j].update,
				 (long long)sb->rec[j].src,
				 (long long)sb->rec[j].dst,
				 sb->rec[j].len, sb->crc);

			if (verbosity < 0) {
				eprintf("%s", msg);
			} else {
				dprintf(verbosity, "%s", msg);
			}
		}
	}
}


/* Start some number of child processes to do the actual work of
 * region swapping.
 */
void
teststart(char *partition)
{
	char lockfile[128];
	char infofile[128];
	char *psrc, *pdst;
	int64_t nsectors;
	info_t *infobase;
	info_t *victim;
	sector_t sb;
	int fdInfo;
	int fdLock;
	int fdPart;
	pid_t pid;
	int rc;

	/* Open the data partition (i.e., the testbed).
	 */
	fdPart = open(partition, O_RDWR);
	if (-1 == fdPart) {
		eprintf("open(%s): %s\n", partition, strerror(errno));
		exit(EX_OSERR);
	}

	/* Get the first record so as to find out the total number
	 * of records involved in the test.
	 */
	getrgn(fdPart, &sb, oseek, (size_t)1, (aiocbptr_t)0);
	chkrgn(&sb, oseek, (size_t)1, (defect_t *)0, (xx_crc32_t *)0);
	nsectors = sb.rec[0].nsectors;

	/* Create the lock file.  We use a lock file rather than the
	 * partition because file locking doesn't work on device special
	 * files.  Each byte in the lock file represents one sector in
	 * the partition file.
	 */
	sprintf(lockfile, "/tmp/%s.XXXXXX", progname);
	fdLock = mkstemp(lockfile);
	if (-1 == fdLock) {
		eprintf("mkstemp(%s): %s\n", lockfile, strerror(errno));
		exit(EX_OSERR);
	}

	if (-1 == ftruncate(fdLock, nsectors)) {
		eprintf("ftruncate(%s): %s\n", lockfile, strerror(errno));
		exit(EX_OSERR);
	}
	(void)close(fdLock);


	/* The info file name is constructed from the partition name with each
	 * forward slash converted to dash.  Existence of this file means that
	 * the program terminated abnormally and the test bed should be
	 * checked.
	 */
	sprintf(infofile, "/var/tmp/%s-", progname);
	pdst = infofile + strlen(infofile);
	for (psrc = partition; *psrc != '\000'; ++psrc) {
		*pdst++ = (*psrc == '/') ? '-' : *psrc;
	}
	*pdst = '\000';

	dprintf(2, "using infofile %s\n", infofile);

	/* The parent removes the lock info file on normal termination.
	 * If the file exists, it means that something bad happened
	 * during the test phase, e.g., the machine crashed or the
	 * parent process died.  In any event, the file needs to be
	 * examined and all in progress transactions rolled back.
	 */
	if (0 == access(infofile, R_OK|W_OK)) {
		eprintf("access(%s): %s\n", infofile, strerror(errno));
		eprintf("Test bed is dirty.  Run `%s -crv %s' to repair.\n",
			progname, partition);
		exit(EX_DATAERR);
	}

	/* Create the shared info file, which is a table of range locks
	 * currently held by each child processes.
	 */
	fdInfo = open(infofile, O_RDWR|O_CREAT, 0600);
	if (-1 == fdInfo) {
		eprintf("open(%s): %s", infofile, strerror(errno));
		exit(EX_OSERR);
	}

	if (-1 == ftruncate(fdInfo, (nprocs + 1) * sizeof(*infobase))) {
		eprintf("ftruncate(%s): %s\n", infofile, strerror(errno));
		exit(EX_OSERR);
	}

#if HAVE_MMAP
	infobase = (info_t *)mmap((void *)0, (nprocs + 1) * sizeof(*infobase),
				  PROT_READ|PROT_WRITE, MAP_SHARED, fdInfo, 0);
	if (infobase == MAP_FAILED) {
		eprintf("mmap: %s", strerror(errno));
		exit(EX_OSERR);
	}
	bzero((char *)infobase, (nprocs + 1) * sizeof(*infobase));
	infobase[nprocs].pid = -1;	/* End of table sentinel */
	victim = infobase;
#else
#error TODO - this implementation does not support mmap
#endif /* HAVE_MMAP */


#if HAVE_ALARM
	(void)signal(SIGHUP, sigHandler);
	(void)signal(SIGINT, sigHandler);
	(void)signal(SIGPIPE, sigHandler);
	(void)signal(SIGXCPU, sigHandler);
	(void)signal(SIGXFSZ, sigHandler);
	(void)signal(SIGVTALRM, sigHandler);
	(void)signal(SIGPROF, sigHandler);
	(void)signal(SIGUSR1, sigHandler);
	(void)signal(SIGUSR2, sigHandler);
	(void)signal(SIGTERM, sigHandler);
#if defined(AIX4) || defined(AIX5)
	(void)signal(SIGDANGER, sigHandler);
#endif
#else
#error TODO - this implementation does not support alarm
#endif /* HAVE_ALARM */

	rc = setpgid(0, getpid());
	if (rc) {
		eprintf("setpgrp: %s\n", strerror(errno));
	}

spawn:
	while (victim->pid == 0) {
		pid = fork();
		switch (pid) {
		case -1:
			eprintf("fork: %s\n", strerror(errno));
			sleep(3);
			continue;

		case 0:
			victim->pid = getpid();

			/* Each child needs his own private file offsets.
			 */
			(void)close(fdPart);
			fdPart = open(partition, O_RDWR);
			if (-1 == fdPart) {
				eprintf("open(%s): %s\n",
					partition, strerror(errno));
				_exit(EX_OSERR);
			}

			fdLock = open(lockfile, O_RDWR);
			if (-1 == fdLock) {
				eprintf("open(%s): %s\n",
					lockfile, strerror(errno));
				_exit(EX_OSERR);
			}

			if (fAIO) {
#if HAVE_AIO
				testaio(infobase, fdInfo,
					fdLock, fdPart, nsectors);
#endif /* HAVE_AIO */
				_exit(0);
			}

			test(infobase, fdInfo, fdLock, fdPart, nsectors);
			_exit(0);

		default:
			++victim;
			break;
		}
	}

	while (1) {
		unsigned int status;

		if (fSig) {
			rc = killpg(0, SIGTERM);
			if (rc) {
				eprintf("killpg: %s\n", strerror(errno));
			}

			fSig = 0;
		}

		pid = wait((int *)&status);
		if (-1 == pid) {
			if (EINTR == errno) {
				continue;
			} else if (ECHILD == errno) {
				break;
			}
			eprintf("wait: %s\n", strerror(errno));
			exit(EX_OSERR);
		}

		/* Find this child's entry in the info table.
		 */
		for (victim = infobase; victim->pid != pid; ++victim) {
			/* Do nothing */
		}
		assert(victim->pid == pid);

		if (WIFSTOPPED(status)) {
			dprintf(1, "stopped(%d) ", WSTOPSIG(status));
		} else if (WIFEXITED(status) && WEXITSTATUS(status)) {
			dprintf(1, "exited(%d) ", WEXITSTATUS(status));
		} else if (WIFSIGNALED(status)) {
			dprintf(1, "signaled(%d) ", WTERMSIG(status));
		}

		/* It may be necessary to roll back any changes if
		 * the child terminated abnomally.
		 */
		if ((WIFEXITED(status) && WEXITSTATUS(status)) ||
		    WIFSIGNALED(status)) {
			cleanlocks(victim, fdInfo, fdPart, nsectors);
			victim->pid = 0;
			goto spawn;
		}

		victim->pid = 0;
	}

	dprintf(2, "removing infofile %s\n", infofile);
	(void)unlink(infofile);

	dprintf(2, "removing lockfile %s\n", lockfile);
	(void)unlink(lockfile);
}


/* Lock the info file and scan all the lock ranges for conflicts.  Conflicts
 * arise due to locks stranded by children who suffered traumatic deaths
 * (e.g., kill -9).  The parent process will eventually clean up the info
 * table and rollback any half baked swaps.
 */
int
itlock(int fdInfo, info_t *infobase, pid_t pid,
       struct flock *sw, struct flock *dw)
{
	struct flock *sh, *dh;			/* src held, dst held */
	struct flock lk;
	info_t *me;
	int rc;

	/* Lock the entire info file (i.e., the shared memory file).
	 */
	lk.l_start = 0;
	lk.l_len = 0;
	lk.l_pid = 0;
	lk.l_whence = SEEK_SET;
	lk.l_type = F_WRLCK;
	if (-1 == fcntl(fdInfo, F_SETLKW, &lk)) {
		eprintf("chklocks(%d): fcntl(%d, F_SETLK): %s\n",
			pid, fdInfo, strerror(errno));
		exit(EX_OSERR);
	}

	/* Check for conflicts.  Conflicts arise due to child processes
	 * that exited (abnormally) without releasing their locked ranges.
	 */
	rc = EBUSY;
	for (me = 0; infobase->pid != -1; ++infobase) {
		if (infobase->pid == pid) {
			me = infobase;
			continue;
		} else if (!infobase->locked) {
			continue;
		}

		sh = &infobase->src;
		dh = &infobase->dst;

		if (!((sw->l_start + sw->l_len <= sh->l_start) ||
		      (sw->l_start >= sh->l_start + sh->l_len))) {
			goto release;
		}
		if (!((sw->l_start + sw->l_len <= dh->l_start) ||
		      (sw->l_start >= dh->l_start + dh->l_len))) {
			goto release;
		}
		if (!((dw->l_start + dw->l_len <= sh->l_start) ||
		      (dw->l_start >= sh->l_start + sh->l_len))) {
			goto release;
		}
		if (!((dw->l_start + dw->l_len <= dh->l_start) ||
		      (dw->l_start >= dh->l_start + dh->l_len))) {
			goto release;
		}
	}

	/* Update the lock ranges and set "locked" to one, thereby
	 * reserving these ranges even if the fcntl locks are lost
	 * due to the child exiting.
	 */
	assert(me && !me->locked);
	me->src = *sw;
	me->dst = *dw;
	me->locked = 1;
	rc = 0;

release:
	lk.l_type = F_UNLCK;
	if (-1 == fcntl(fdInfo, F_SETLKW, &lk)) {
		eprintf("chklocks(%d): fcntl(F_UNLCK): %s\n",
			pid, strerror(errno));
		exit(EX_OSERR);
	}

	return rc;
}


void
itunlock(int fdInfo, info_t *me)
{
	struct flock lk;

	assert(me);

	lk.l_start = 0;
	lk.l_len = 0;
	lk.l_pid = 0;
	lk.l_whence = SEEK_SET;

	/* Release info table record lock.
	 */
	lk.l_type = F_WRLCK;
	if (-1 == fcntl(fdInfo, F_SETLKW, &lk)) {
		eprintf("test(%d): fcntl(info, F_WRLCK): %s\n",
			getpid(), strerror(errno));
		exit(EX_OSERR);
	}

	assert(me->locked);
	me->locked = 0;

	lk.l_type = F_UNLCK;
	if (-1 == fcntl(fdInfo, F_SETLKW, &lk)) {
		eprintf("test(%d): fcntl(info, F_UNLCK): %s\n",
			getpid(), strerror(errno));
		exit(EX_OSERR);
	}
}


/* cleanlocks is called to clean up stale locks left by a process
 * that exited abnormally.  If the process died while holding locks,
 * then me->locked will be `true' and we have to examine the disk
 * regions covered by the child's range locks (src and dst), and
 * roll back records involved in an interrupted swap (if any).
 */
int
cleanlocks(info_t *me, int fdInfo, int fdPart, int64_t nsectors)
{
	sector_t *sb;
	defect_t *d;

	if (!me->locked) {
		return 0;
	}

	if ((me->src.l_len == 0) || (me->dst.l_len == 0)) {
		dprintf(1, "cleanlocks: len = 0 %ld %ld\n",
			me->src.l_len, me->dst.l_len);
		me->locked = 0;
		return 0;
	}
	assert(me->src.l_len == me->dst.l_len);

	/* At this point we know the child had some range locks,
	 * so we need to see if there are any records that need
	 * to be rolled back.
	 */
	dprintf(1, "cleanlocks: child %d has range locks...\n", me->pid);
	dprintf(2, "cleanlocks: src: l_start=%lld l_len=%d\n",
		(long long)me->src.l_start, me->src.l_len);
	dprintf(2, "cleanlocks: dst: l_start=%lld l_len=%d\n",
		(long long)me->dst.l_start, me->dst.l_len);

	sb = malloc(maxswapsectors * sizeof(*sb));
	if (!sb) {
		eprintf("cleanlocks: out of memory\n");
		return 0;
	}

	d = defect_alloc(nsectors + oseek);
	if (!d) {
		eprintf("cleanlocks: out of memory\n");
		free(sb);
		return 0;
	}

	/* Get the regions involved in the swap.
	 */
	getrgn(fdPart, sb, me->src.l_start, me->src.l_len, (aiocbptr_t)0);
	chkrgn(sb, me->src.l_start, me->src.l_len, d, (xx_crc32_t *)0);

	getrgn(fdPart, sb, me->dst.l_start, me->dst.l_len, (aiocbptr_t)0);
	chkrgn(sb, me->dst.l_start, me->dst.l_len, d, (xx_crc32_t *)0);

	repair(fdPart, d);
	me->locked = 0;

	free(sb);
	defect_free(d);

	dprintf(1, "cleanlocks: child %d cleaned\n", me->pid);

	return 0;
}


/* Continuously select two arbitrary contiguous regions and swap them.
 */
void
test(info_t *infobase, int fdInfo, int fdLock, int fdPart, int64_t nsectors)
{
	time_t starttime, now;
	struct flock src, dst;
	sector_t *sb1, *sb2;
	int64_t nblkswr;
	u_int nswaps;
	size_t len;
	info_t *me;
	pid_t pid;

	pid = getpid();

	/* Add pid to the seed so that the child processes
	 * don't all generate the same sequence.
	 */
	starttime = time((time_t *)0);
	srandom(starttime+pid);
	nswaps = 0;

	/* Allocate two buffers into which to hold the data to be swapped.
	 */
	sb1 = malloc(maxswapsectors * sizeof(*sb1));
	sb2 = malloc(maxswapsectors * sizeof(*sb2));
	if (!(sb1 && sb2)) {
		eprintf("(%d): insufficient memory for %lld swapsector "
			"buffers\n", pid, maxswapsectors);
		exit(EX_OSERR);
	}
	bzero(sb1, sizeof(*sb1) * maxswapsectors);
	bzero(sb2, sizeof(*sb2) * maxswapsectors);

	if ((void(*)(int))-1 == signal(SIGALRM, sigAlarm)) {
		eprintf("(%d): signal: %s\n", pid, strerror(errno));
		exit(EX_OSERR);
	}

	/* Find my entry in the info table.
	 */
	for (me = infobase; me->pid != pid; ++me) {
		/* Do nothing */
	}
	assert(me->pid == pid);

	nblkswr = 0;
	while (!fSig) {
		len = (random() % maxswapsectors) + 1;

		/* Get write locks.
		 */
		src.l_start = (random() % (nsectors - len - oseek)) + oseek;
		src.l_len = len;
		src.l_pid = 0;
		src.l_type = F_WRLCK;
		src.l_whence = SEEK_SET;

		/* Set a watchdog so as to break deadlocks.
		 */
		fSigAlarm = 0;
		alarm(30);

		/* Try for an exclusive lock over the src region,
		 * blocking if necessary (F_SETLKW vs F_SETLK).
		 */
		if (-1 == fcntl(fdLock, F_SETLKW, &src)) {
			alarm(0);
			if (EINTR == errno && fSig) {
				continue;
			} else if (EINTR == errno && fSigAlarm) {
				eprintf("test(%d): fcntl: deadlock?\n", pid);
				continue;
			}
			eprintf("test(%d) fcntl(srclock, F_WRLCK): %s\n",
				pid, strerror(errno));
			exit(EX_OSERR);
		}
		alarm(0);

		/* Try for an exclusive lock over the dst region,
		 * but do not block (F_SETLK).
		 */
		while (1) {
			dst.l_start =
				(random() % (nsectors - len - oseek)) + oseek;
			dst.l_len = len;
			dst.l_pid = 0;
			dst.l_type = F_WRLCK;
			dst.l_whence = SEEK_SET;

			/* If the dst region overlaps the src region
			 * then reselect the dst region.
			 */
			if (!((dst.l_start > src.l_start + len) ||
			      (dst.l_start + len < src.l_start))) {
				continue;
			}

			if (-1 == fcntl(fdLock, F_SETLK, &dst)) {
				if (EAGAIN == errno) {
					continue;
				} else if (EINTR == errno) {
					continue;
#if defined(AIX4) || defined(AIX5)
				} else if (EACCES == errno) {
					continue;
#endif
				}

				eprintf("test(%d): fcntl(dstlock, F_WRLCK): "
					"%s start=%lld len=%d\n",
					pid, strerror(errno),
					(long long)dst.l_start, len);
				exit(EX_OSERR);
			}
			break;
		}

		if (src.l_start < 0 || dst.l_start < 0) {
			eprintf("test(%d): huh?", pid);
			exit(69);
		}

		/* Acquire info table lock.
		 */
		if (0 != itlock(fdInfo, infobase, pid, &src, &dst)) {
			dprintf(1, "test(%d): stranded lock detected\n", pid);
			goto release;
		}

		dprintf(3, "test(%d): swapping [%lld %lld %d]\n",
			pid, (long long)src.l_start, (long long)dst.l_start, len);

		/* Read and check the data from the selected regions.
		 */
		getrgn(fdPart, sb1, src.l_start, len, (aiocbptr_t)0);
		chkrgn(sb1, src.l_start, len, (defect_t *)0, (xx_crc32_t *)0);

		getrgn(fdPart, sb2, dst.l_start, len, (aiocbptr_t)0);
		chkrgn(sb2, dst.l_start, len, (defect_t *)0, (xx_crc32_t *)0);

		/* Save the rollback records.
		 */
		saverollback(sb1, sb2, len);

		/* Update all records.
		 */
		now = time((time_t *)0);
		updatergn(sb1, src.l_start, dst.l_start, len, pid, now);
		updatergn(sb2, dst.l_start, src.l_start, len, pid, now);

		/* Swap and write the regions.  Always write the range
		 * with the lowered disk block address first to make
		 * rollback easier.
		 */
		if (fExec) {
			if (src.l_start < dst.l_start) {
				putrgn(fdPart, sb2, src.l_start, len,
				       (aiocbptr_t)0);
				putrgn(fdPart, sb1, dst.l_start, len,
				       (aiocbptr_t)0);
			} else {
				putrgn(fdPart, sb1, dst.l_start, len,
				       (aiocbptr_t)0);
				putrgn(fdPart, sb2, src.l_start, len,
				       (aiocbptr_t)0);
			}
		}

		/* Release info table record lock.
		 */
		itunlock(fdInfo, me);

		/* Release file range locks.
		 */
	release:
		dst.l_type = F_UNLCK;
		if (-1 == fcntl(fdLock, F_SETLKW, &dst)) {
			eprintf("test(%d): fcntl(dstlock, F_UNLCK): %s\n",
				pid, strerror(errno));
			exit(EX_OSERR);
		}

		src.l_type = F_UNLCK;
		if (-1 == fcntl(fdLock, F_SETLKW, &src)) {
			eprintf("test(%d): fcntl(srclock, F_UNLCK): %s\n",
				pid, strerror(errno));
			exit(EX_OSERR);
		}

		++nswaps;
		if ((maxiterations > 0) && (nswaps >= maxiterations)) {
			break;
		}

		/* Add the number of blocks just written to the running
		 * total (nblkswr).  If the number of blocks written per
		 * second thus far (nblkswrps) exceeds the maximum desired
		 * number of blocks written per second (maxblksps), then
		 * sleep long enough to try and keep the rates as
		 * close as possible.
		 */
		nblkswr += len * 2;

		if ((maxblksps > 0) && (now > starttime)) {
			static int64_t nblkswr_saved = 0;
			static u_long delay = 0;

			if (nblkswr_saved < nblkswr) {
				long nblkswrps = nblkswr / (now - starttime);
				u_long adjust;

				nblkswr_saved = nblkswr;
				adjust = nblkswrps * 100000 / maxblksps;

				dprintf(3, "(%d): nblkswrps=%ld nblkswr=%lld "
					"delay=%lu\n", pid,
					nblkswrps, nblkswr, delay);

				if (nblkswrps > maxblksps) {
					delay += adjust;
					usleep(delay);
				} else if (delay > adjust) {
					delay -= adjust;
				}
			}
		}
	}

	free(sb1);
	free(sb2);

	now = time((time_t *)0);
	dprintf(1, "(%d): nswaps=%u seconds=%d swaps/sec=%.1f nblkswr=%lld "
		"blkswr/sec=%.1f\n", pid, nswaps, now - starttime,
		(float)nswaps / (now - starttime),
		nblkswr, (float)nblkswr / (now - starttime));
}


#if HAVE_AIO
void
testaio(info_t *infobase, int fdInfo, int fdLock, int fdPart, int64_t nsectors)
{
	time_t starttime, now;
	struct aiocb **arwaio;
	int maxtries = 8;
	int64_t nblkswr;
	xaio_t **axaio;
	u_int nswaps;
	size_t nrgns;
	int rwidx;
	pid_t pid;
	int nrw;

	/* Allocate the aiocb pointer array.  This array holds
	 * pointers to aiocbs that are in flight.
	 */
	arwaio = malloc(naioreqs * sizeof(*arwaio));
	if (!arwaio) {
		eprintf("%d: insufficient memory for %d aio requests\n",
			naioreqs);
		exit(EX_OSERR);
	}
	bzero(arwaio, naioreqs * sizeof(*arwaio));

	/* Allocate the xaio pointer array.  This array hold
	 * pointers to xaios for which there is an active
	 * swap in progress.
	 */
	axaio = malloc(naioreqs * sizeof(*axaio));
	if (!axaio) {
		eprintf("%d: insufficient memory for %d aio requests\n",
			naioreqs);
		exit(EX_OSERR);
	}
	bzero(axaio, naioreqs * sizeof(*axaio));

	pid = getpid();

	starttime = time((time_t *)0);
	srandom(starttime+pid);
	nswaps = 0;

	if ((void(*)(int))-1 == signal(SIGALRM, sigAlarm)) {
		eprintf("%d: signal: %s\n", strerror(errno));
		exit(EX_OSERR);
	}

	nrw = 0;
	rwidx = 0;
	nblkswr = 0;

	while (1) {
		xaio_t *x1, *x2;
		int tries;
		int gbrc;
		int rc;
		int i;

		nrgns = (random() % maxswapsectors) + 1;

		/* Create two xaio control buffers, one for
		 * each region that we are going to swap.
		 */
		assert(nrgns <= maxswapsectors);
		gbrc = getaiobuf(nrgns, fdLock, &x1, &x2);
		if (gbrc) {
			goto reaprw;
		}

		x1->read = 1;
		x2->read = 1;

		/* Try for an exclusive lock over the first region,
		 * but do not block (F_SETLK).
		 */
		x1->idx = getidx(naioreqs, axaio, &rwidx);
		for (tries = 0; tries < maxtries; ++tries) {
			x1->lk.l_start =
				random() % (nsectors - nrgns - oseek) + oseek;
			x1->lk.l_len = nrgns;
			x1->lk.l_pid = 0;
			x1->lk.l_type = F_WRLCK;
			x1->lk.l_whence = SEEK_SET;

			/* Check that this lock range doesn't overlap
			 * any other request that is in progress.
			 *
			 * TODO: This code spins if naioreqs is too
			 * large for the testbed.
			 */
			if (chkoverlap(naioreqs, axaio, x1)) {
				dprintf(3, "testaio(%d): overlapping "
					"lock rgns start=%lld\n",
					pid, (long long)x1->lk.l_start);
				continue;
			}

			/* Get an exclusive range lock to prevent
			 * concurrent access to this region by any
			 * other process working in this test bed.
			 */
			if (-1 == fcntl(fdLock, F_SETLK, &x1->lk)) {
				dprintf(4, "testaio(%d): fcntl(F_SETLK): %s\n",
					pid, strerror(errno));
				sleep(1);

				if (EAGAIN == errno) {
					continue;
				} else if (EINTR == errno) {
					continue;
#if defined(AIX4) || defined(AIX5)
				} else if (EACCES == errno) {
					continue;
#endif
				}

				eprintf("testaio(%d): fcntl(F_SETLK): %s\n",
					pid, strerror(errno));
				exit(EX_OSERR);
			}
			break;
		}

		if (tries < maxtries) {
			axaio[x1->idx] = x1;
			arwaio[x1->idx] = &x1->aio;
		} else {
			dprintf(3, "testaio(%d): couldn't lock first "
				"region\n", pid);

			x1->read = 0;
			x2->read = 0;
			freeaiobuf(x1, x2);
			goto reaprw;
		}

		/* Try for an exclusive lock over the second region,
		 * but do not block (F_SETLK).
		 */
		x2->idx = getidx(naioreqs, axaio, &rwidx);
		for (tries = 0; tries < maxtries; ++tries) {
			x2->lk.l_start =
				random() % (nsectors - nrgns - oseek) + oseek;
			x2->lk.l_len = nrgns;
			x2->lk.l_pid = 0;
			x2->lk.l_type = F_WRLCK;
			x2->lk.l_whence = SEEK_SET;

			/* Check that this lock range doesn't overlap
			 * any other request that is in progress.
			 *
			 * TODO: This code spins if naioreqs is too
			 * large for the testbed.
			 */
			if (chkoverlap(naioreqs, axaio, x2)) {
				dprintf(3, "testaio(%d): overlapping "
					"lock rgns start=%lld\n",
					pid, (long long)x2->lk.l_start);
				continue;
			}

			/* Get an exclusive range lock to prevent
			 * concurrent access to this region by any
			 * other process working in this test bed.
			 */
			if (-1 == fcntl(fdLock, F_SETLK, &x2->lk)) {
				dprintf(4, "testaio(%d): F_SETLK: %s\n",
					pid, strerror(errno));
				sleep(1);

				if (EAGAIN == errno) {
					continue;
				} else if (EINTR == errno) {
					continue;
#if defined(AIX4) || defined(AIX5)
				} else if (EACCES == errno) {
					continue;
#endif
				}

				eprintf("testaio(%d): fcntl(F_WRLCK): "
					"%s start=%lld nrgns=%ld\n",
					pid, strerror(errno),
					(long long)x1->lk.l_start, nrgns);
				exit(EX_OSERR);
			}
			break;
		}

		if (tries < maxtries) {
			axaio[x2->idx] = x2;
			arwaio[x2->idx] = &x2->aio;

			dprintf(3, "testaio(%d): aio read [%lld %lld %ld]\n",
				pid, (long long)x1->lk.l_start,
				(long long)x2->lk.l_start, x1->nrgns);

			/* Read the data from the selected regions.
			 */
			getrgn(fdPart, x1->sb, x1->lk.l_start,
			       x1->nrgns, &x1->aio);
			getrgn(fdPart, x2->sb, x2->lk.l_start,
			       x2->nrgns, &x2->aio);

			nrw += 2;
		} else {
			x1->lk.l_type = F_UNLCK;
			if (-1 == fcntl(fdLock, F_SETLKW, &x1->lk)) {
				eprintf("testaio(%d): fcntl(F_UNLCK): "
					"%s\n", pid, strerror(errno));
				exit(EX_OSERR);
			}

			dprintf(3, "testaio(%d): couldn't lock second "
				"region\n", pid);

			x1->read = 0;
			x2->read = 0;
			axaio[x1->idx] = 0;
			arwaio[x1->idx] = 0;
			freeaiobuf(x1, x2);
			goto reaprw;
		}


	  reaprw:
		/* Here we scan the array of requests to reap
		 * read requests that have completed.
		 */
		for (i = 0; i < naioreqs; ++i) {
			x1 = axaio[i];

			if (!x1 || !x1->read || !arwaio[i]) {
				continue;
			}
			assert(!x1->done);

			rc = AIO_ERROR(&x1->aio);
			if (rc) {
				if (rc != EINPROGRESS) {
					eprintf("testaio(%d): aio_error: %s\n",
						pid, strerror(rc));
					exit(EX_OSERR);
				}
				continue;
			}

			x1->done = 1;
			x1->read = 0;
			x2 = x1->sibling;
			arwaio[i] = 0;

			dprintf(3, "testaio(%d): read done [%lld %lld %ld]\n",
				pid, (long long)x1->lk.l_start,
				(long long)x2->lk.l_start, x1->nrgns);
				
			chkrgn(x1->sb, x1->lk.l_start,
			       x1->nrgns, (defect_t *)0, (xx_crc32_t *)0);


			/* If the sibling is done, then both reads have
			 * completed.  We can now initiate the write/swap
			 * phase of the transaction.
			 */
			if (x2->done) {
				x1->done = x2->done = 0;

				/* Save the rollback records.
				 */
				saverollback(x1->sb, x2->sb, x1->nrgns);

				arwaio[x1->idx] = &x1->aio;
				arwaio[x2->idx] = &x2->aio;

				dprintf(3, "testaio(%d): aio write "
					"[%lld %lld %ld]\n",
					pid, (long long)x1->lk.l_start,
					(long long)x2->lk.l_start, x1->nrgns);

				now = time((time_t *)0);
				updatergn(x1->sb, x1->lk.l_start,
					  x2->lk.l_start, x1->nrgns, pid, now);
				updatergn(x2->sb, x2->lk.l_start,
					  x1->lk.l_start, x2->nrgns, pid, now);

				putrgn(fdPart, x1->sb, x2->lk.l_start,
				       x1->nrgns, &x1->aio);
				putrgn(fdPart, x2->sb, x1->lk.l_start,
				       x2->nrgns, &x2->aio);

				nblkswr += x1->nrgns + x2->nrgns;


			}
		}


		/* Here we scan the array of requests to reap
		 * write requests that have completed.
		 */
		for (i = 0; i < naioreqs; ++i) {
			x1 = axaio[i];

			if (!x1 || x1->read || !arwaio[i]) {
				continue;
			}
			assert(!x1->done);

			rc = AIO_ERROR(&x1->aio);
			if (rc) {
				if (rc != EINPROGRESS) {
					eprintf("testaio(%d): aio_error: %s\n",
						pid, strerror(rc));
					exit(EX_OSERR);
				}
				continue;
			}

			x1->done = 1;
			x2 = x1->sibling;
			arwaio[i] = 0;

			dprintf(3, "testaio(%d): write done [%lld %lld %ld]\n",
				pid, (long long)x1->lk.l_start,
				(long long)x2->lk.l_start, x1->nrgns);

			/* Release range lock.
			 */
			x1->lk.l_type = F_UNLCK;
			if (-1 == fcntl(fdLock, F_SETLKW, &x1->lk)) {
				eprintf("testaio(%d): fcntl(F_UNLCK): "
					"%s\n", pid, strerror(errno));
				exit(EX_OSERR);
			}

			/* If the sibling is done, then both writes
			 * have completed and the swap is complete.
			 */
			if (x2->done) {
				axaio[x1->idx] = 0;
				axaio[x2->idx] = 0;
				freeaiobuf(x1, x2);

				nrw -= 2;
				++nswaps;
				if ((maxiterations > 0) &&
				    (--maxiterations == 0)) {
					++fSig;
				}
			}
		}

		/* If we caught a signal to terminate, then we must
		 * wait for all the in flight I/O to finish.
		 */
		if (fSig) {
			static int gavemsg = 0;

			if (!nrw) {
				break;
			} else if (!gavemsg) {
				dprintf(1, "testaio(%d): flushing %d "
					"buffers...\n", pid, nrw);
				fflush(stdout);
				gavemsg = 1;
			}
		}

		/* Here we must stay in the reaper loop until space
		 * frees up in the aio request tracking table (axio[]).
		 */
		if ((naioreqs - nrw < 2) || gbrc || fSig) {
			dprintf(3, "testaio(%d): nrw=%d gbrc=%d fsig=%d\n",
				pid, nrw, gbrc, fSig);
			if (nrw > 0) {
				rc = AIO_SUSPEND(arwaio, naioreqs);
				if (rc == -1) {
					dprintf(4, "testaio(%d): aio_suspend: "
						"%s nrw=%d\n",
						pid, strerror(errno), nrw);
					sleep(1);
				}
			}
			goto reaprw;
		}

		/* Add the number of blocks just written to the running
		 * total (nblkswr).  If the number of blocks written per
		 * second thus far (nblkswrps) exceeds the maximum desired
		 * number of blocks written per second (maxblksps), then
		 * sleep long enough to try and keep the rates as
		 * close as possible.
		 */
		if ((maxblksps > 0) && (now > starttime)) {
			static int64_t nblkswr_saved = 0;
			static u_long delay = 0;

			if (nblkswr_saved < nblkswr) {
				long nblkswrps = nblkswr / (now - starttime);
				u_long adjust;

				nblkswr_saved = nblkswr;
				adjust = nblkswrps * 100000 / maxblksps;

				dprintf(3, "(%d): nblkswrps=%ld nblkswr=%lld "
					"delay=%lu\n", pid,
					nblkswrps, nblkswr, delay);

				if (nblkswrps > maxblksps) {
					delay += adjust;
					usleep(delay);
				} else if (delay > adjust) {
					delay -= adjust;
				}
			}
		}
	}

	now = time((time_t *)0);
	dprintf(1, "(%d): nswaps=%u seconds=%d swaps/sec=%.1f nblkswr=%lld "
		"blkswr/sec=%.1f\n", pid, nswaps, now - starttime,
		(float)nswaps / (now - starttime),
		nblkswr, (float)nblkswr / (now - starttime));
}
#endif /* HAVE_AIO */


/* Update region prepares each record in the given range
 * to be written to disk.
 */
void
updatergn(sector_t *sb, int64_t src, int64_t dst,
	  size_t len, pid_t pid, time_t time)
{
	int i;

	for (i = 0; i < len; ++i, ++sb) {
		record_t *r = &sb->rec[0];

		r->pid = pid;
		r->update = time;
		r->src = src + i;
		r->dst = dst + i;
		r->len = len;
		++r->nswaps;

		sb->kscratch = 0;
		sb->crc = 0;
		sb->crc = xxcrc32((u_char *)sb, sizeof(*sb));
	}
}


/* For each record in the given range, check region calculates
 * the crc and compares it to the stored crc in effort to
 * validate that the record has not been corrupted.
 */
void
chkrgn(sector_t *sb, int64_t offset, size_t len, defect_t *d,
       xx_crc32_t *signature)
{
	int i;

	/* Check the integrity of the data read.
	 */
	for (i = 0; i < len; ++i, ++sb) {
		record_t *r = &sb->rec[0];
		xx_crc32_t crc;

		/* The kernel scratch area is used for in-kernel
		 * validation (sequence number validation, etc, ...
		 * We must set it to zero before computing the crc.
		 */
		sb->kscratch = 0;

		crc = sb->crc;
		sb->crc = 0;
		sb->crc = xxcrc32((u_char *)sb, sizeof(*sb));
		if (crc != sb->crc) {
			eprintf("ckhrgn: crc mismatch at sector %lld\n",
				offset + i);
			exit(EX_DATAERR);
		}

		if (signature) {
			*signature = (*signature << 8) ^ sb->crc;
		}

		if (r->dst != offset + i) {
			eprintf("ckhrgn: dst block mismatch (%lld != %lld)\n",
				r->dst, offset + i);
			dumprgn(sb, 0, 1, -1);
		}

		if (d) {
			if (r->id == (offset + i)) {
				++d->inplace;
			}

			if (d->x[r->id] == -1) {
				d->x[r->id] = offset + i;
			} else if (d->y[r->id] == -1) {
				d->y[r->id] = offset + i;
			} else {
				eprintf("chkrgn: sector %lld occurred more "
					"than twice\n", offset + i);
				/*exit(EX_DATAERR);*/
			}
		}
	}
}


/* Called prior to a swap write.  Saves the previous contents of rec[0]
 * into rec[1] so that rec[0] can be rolled back in the case of
 * catastrophic error (e.g., system crash, kill -9, etc, ...)
 */
void
saverollback(sector_t *x, sector_t *y, size_t nrgns)
{
	while (nrgns-- > 0) {
		x[nrgns].rec[1] = y[nrgns].rec[0];
		y[nrgns].rec[1] = x[nrgns].rec[0];
	}
}


/* Note that we received a signal.
 */
RETSIGTYPE
sigHandler(int sig)
{
	++fSig;
}


/* Note that the watchdog timer expired.
 */
RETSIGTYPE
sigAlarm(int sig)
{
	++fSigAlarm;
}


/* Debug print.
 */
void
dprintf(int lvl, char *fmt, ...)
{
	char msg[128];
	va_list ap;

	if (verbosity >= lvl) {
		va_start(ap, fmt);
		vsnprintf(msg, sizeof(msg), fmt, ap);
		va_end(ap);

		fputs(msg, stdout);
	}
}


/* Error print.
 */
void
eprintf(char *fmt, ...)
{
	char msg[128];
	va_list ap;

	sprintf(msg, "%s: ", progname);

	va_start(ap, fmt);
	vsnprintf(msg+strlen(msg), sizeof(msg)-strlen(msg), fmt, ap);
	va_end(ap);

	fputs(msg, stderr);
}


#if HAVE_AIO
int
getaiobuf(int nrgns, int fd, xaio_t **x1, xaio_t **x2)
{
	assert(sizeof(xaio_t) <= sizeof(sector_t));

	if (xaiohead) {
		*x1 = xaiohead;
		xaiohead = (*x1)->next;
		if ((*x1)->next) {
			(*x1)->next = (xaio_t *)0;
		}
		assert((*x1)->magic == 0xdeadbeef);
	} else {
		(*x1) = malloc((maxswapsectors+1) * sizeof(sector_t));
		if (!(*x1)) {
			dprintf(3, "getaiobuf: malloc(%ld) failed...\n",
				(nrgns+1) * sizeof(sector_t));
			return ENOMEM;
		}
	}

	bzero(*x1, sizeof(**x1));
	(*x1)->sb = (sector_t *)((char *)(*x1) + sizeof(sector_t));
	(*x1)->nrgns = nrgns;
	(*x1)->magic = 0x01234567;


	if (xaiohead) {
		*x2 = xaiohead;
		xaiohead = (*x2)->next;
		if ((*x2)->next) {
			(*x2)->next = (xaio_t *)0;
		}
		assert((*x2)->magic == 0xdeadbeef);
	} else {
		*x2 = malloc((maxswapsectors+1) * sizeof(sector_t));
		if (!(*x2)) {
			dprintf(3, "getaiobuf: malloc(%ld) failed...\n",
				(nrgns+1) * sizeof(sector_t));
			(*x1)->magic = 0xdeadbeef;
			(*x1)->next = xaiohead;
			xaiohead = (*x1);
			return ENOMEM;
		}
	}

	bzero(*x2, sizeof(**x2));
	(*x2)->sb = (sector_t *)((char *)(*x2) + sizeof(sector_t));
	(*x2)->nrgns = nrgns;
	(*x2)->magic = 0x01234567;

	(*x1)->sibling = *x2;
	(*x2)->sibling = *x1;

	return 0;
}


void
freeaiobuf(xaio_t *xaio1, xaio_t *xaio2)
{
	assert(xaio1->magic == 0x01234567);
	assert(xaio2->magic == 0x01234567);

	xaio1->magic = 0xdeadbeef;
	xaio2->magic = 0xdeadbeef;

	xaio2->next = xaiohead;
	xaio1->next = xaio2;
	xaiohead = xaio1;
}


int
getidx(int n, xaio_t **ap, int *pidx)
{
	int idx = *pidx;
	int i;

	for (i = 0; i < n; ++i) {
		if (!ap[idx]) {
			return (*pidx = idx);
		}
		idx = (idx + 1) % n;
	}

	eprintf("getidx: array is full!\n");

	for (i = 0; i < n; ++i) {
		fprintf(stderr, "%d: %p\n", i, ap[i]);
	}

	sleep(3);

	exit(EX_DATAERR);
}


int
chkoverlap(int n, xaio_t **ap, xaio_t *tgt)
{
	int i;

	for (i = 0; i < n; ++i) {
		xaio_t *x = ap[i];

		if (x) {
			if (x->lk.l_start + x->lk.l_len <
			    tgt->lk.l_start) {
				continue;
			}

			if (tgt->lk.l_start + tgt->lk.l_len <
			    x->lk.l_start) {
				continue;
			}

			return 1;
		}
	}

	return 0;
}
#endif /* HAVE_AIO */



/* The following C code (by Rob Warnock <rpw3@sgi.com>) does CRC-32 in
 * BigEndian/BigEndian byte/bit order.  That is, the data is sent most
 * significant byte first, and each of the bits within a byte is sent most
 * significant bit first, as in FDDI. You will need to twiddle with it to do
 * Ethernet CRC, i.e., BigEndian/LittleEndian byte/bit order. [Left as an
 * exercise for the reader.]

 * The CRCs this code generates agree with the vendor-supplied Verilog models
 * of several of the popular FDDI "MAC" chips.
 */

/* Build auxiliary table for parallel byte-at-a-time CRC-32.
 */
#define CRC32_POLY 0x04c11db7     /* AUTODIN II, Ethernet, & FDDI */

static xx_crc32_t xxcrc32_table[256];

void
xxcrc32_init(void)
{
	xx_crc32_t c;
	int i, j;

	for (i = 0; i < 256; ++i) {
		for (c = i << 24, j = 8; j > 0; --j)
			c = c & 0x80000000 ? (c << 1) ^ CRC32_POLY : (c << 1);
		xxcrc32_table[i] = c;
	}
}


xx_crc32_t
xxcrc32(u_char *buf, int len)
{
	xx_crc32_t crc;
	u_char *p;

	crc = 0xffffffff;       /* preload shift register, per CRC-32 spec */
	for (p = buf; len > 0; ++p, --len) {
		crc = (crc << 8) ^ xxcrc32_table[(crc >> 24) ^ *p];
	}

	return ~crc;            /* transmit complement, per CRC-32 spec */
}


/* Read "len" number of sectors into "sb" starting from "offset".
 */
void
getrgn(int fd, sector_t *sb, int64_t offset, size_t len, aiocbptr_t aio)
{
	ssize_t cc;

	/* Read the data over the locked regions.
	 */
#if HAVE_AIO
	if (aio) {
		int rc;

		bzero(aio, sizeof(*aio));
		aio->aio_offset = offset * sizeof(*sb);
		aio->aio_buf = (char *)sb;
		aio->aio_nbytes = len * sizeof(*sb);

#if defined(AIX4) || defined(AIX5)
		aio->aio_whence = SEEK_SET;
		aio->aio_flag = AIO_SIGNAL;
#endif /* AIX4 || AIX4 */

#ifdef _AIO_AIX_SOURCE
		/* Legacy aio.
		 */
		/* aio->aio_event Not Used */
		rc = aio_read(fd, aio);
#else
		/* Posix aio.
		 */
		aio->aio_lio_opcode = LIO_READ;
		aio->aio_fildes = fd;
		rc = aio_read(aio);
#endif /* _AIO_AIX_SOURCE */

		if (rc == -1) {
			eprintf("getrgn: aio_read: %s\n", strerror(errno));
			exit(EX_OSERR);
		}
		return;
	}
#endif /* HAVE_AIO */

	/* TODO: We should use probably use pread(), if available.
	 */

	if (-1 == lseek(fd, offset * sizeof(*sb), SEEK_SET)) {
		eprintf("getrgn: lseek: %s\n", strerror(errno));
		exit(EX_OSERR);
	}

	cc = read(fd, sb, len * sizeof(*sb));
	if (-1 == cc) {
		eprintf("getrgn: read: %s\n", strerror(errno));
		exit(EX_IOERR);
	} else if (cc != len * sizeof(*sb)) {
		eprintf("getrgn: read: tried to read %ld bytes at offset "
			"%ld but only %ld bytes were read\n", (long)cc,
			(long)offset * sizeof(*sb),
			(long)len * sizeof(*sb));
		assert(0);
		exit(EX_IOERR);
	}
}


/* Write the given records to disk.
 */
void
putrgn(int fd, sector_t *sb, int64_t offset, size_t len, aiocbptr_t aio)
{
	off64_t off;
	ssize_t cc;

#if HAVE_AIO
	if (aio) {
		int rc;

		bzero(aio, sizeof(*aio));
		aio->aio_offset = offset * sizeof(*sb);
		aio->aio_buf = (char *)sb;
		aio->aio_nbytes = len * sizeof(*sb);

#if defined(AIX4) || defined(AIX5)
		aio->aio_flag = AIO_SIGNAL;
		aio->aio_whence = SEEK_SET;
#endif /* AIX4 || AIX4 */

#ifdef _AIO_AIX_SOURCE
		/* Legacy aio.
		 */
		/* aio->aio_event Not Used */
		rc = aio_write(fd, aio);
#else
		/* Posix aio.
		 */
		aio->aio_lio_opcode = LIO_WRITE;
		aio->aio_fildes = fd;
		rc = aio_write(aio);
#endif /* _AIO_AIX_SOURCE */

		if (rc == -1) {
			eprintf("putrgn: aio_write: %s\n", strerror(errno));
			exit(EX_OSERR);
		}
		return;
	}
#endif /* HAVE_AIO */

	/* TODO: We should use probably use pwrite(), if available.
	 */

	off = lseek(fd, offset * sizeof(*sb), SEEK_SET);
	if (off != (offset * sizeof(*sb))) {
		if (-1 == off) {
			eprintf("putrgn: lseek: %s\n",
				strerror(errno));
		} else {
			eprintf("putrgn: lseek: invalid offset %d\n",
				off);
		}
		exit(EX_OSERR);
	}

	cc = write(fd, sb, len * sizeof(*sb));
	if (-1 == cc) {
		eprintf("putrgn: write: %s\n", strerror(errno));
		exit(EX_IOERR);
	} else if (cc != (len * sizeof(*sb))) {
		eprintf("putrgn: write: tried to write %ld bytes at offset "
			"%ld but only %ld bytes were written\n", (long)cc,
			(long)offset * sizeof(*sb),
			(long)len * sizeof(*sb));
		assert(0);
		exit(EX_IOERR);
	}
}


